SUBROUTINE Evaluate3D( freq, k, PhiR, PhiS, M, maxM, IElementSource, xs, ys, theta, Ntheta, RminM, RmaxM, Nr, MLimit, P )

  ! Computes 3-D pressure field using adiabatic mode theory           
  ! Normalized to pressure of point source at 1 meter                 
  ! Note RminM must be zero
  ! mike porter, 1987

  USE ElementMod 
  USE beampatternMod
  USE interpMod
  IMPLICIT NONE
  INTEGER, PARAMETER    :: KBarFile = 55, ZBarFile = 56
  REAL,    PARAMETER    :: pi = 3.141592, DegRad = pi / 180.0, RadDeg = 180.0 / pi
  COMPLEX, PARAMETER    :: i  = ( 0.0, 1.0 )
  INTEGER, INTENT( IN ) :: M( * ), MLimit, maxM            ! number of modes, limit on number of modes to propagate
  INTEGER, INTENT( IN ) :: Nr, Ntheta, IElementSource      ! number of receiver ranges and bearing lines
  REAL,    INTENT( IN ) :: xs, ys, freq                    ! source coordinate and frequency
  REAL,    INTENT( IN ) :: theta( Ntheta )                 ! bearing angles for receivers
  COMPLEX, INTENT( IN ) :: k( maxM, * )                    ! wavenumbers
  COMPLEX, INTENT( OUT) :: P( Ntheta, Nr )                 ! pressure field
  INTEGER               :: iElement, ir, Itheta, L, MProp, NewElement, Outside
  REAL                  :: tsx, tsy                        ! tangent vector from source
  REAL                  :: RminM, RmaxM                    ! minimum and maximum receiver ranges in m
  REAL                  :: alpha, delta_r, Rin, Rout, RM
  COMPLEX               :: PhiR( maxM, * ), PhiIn( maxM ), PhiOut( maxM ), PhiS( maxM, * ), const( maxM ), PhiInt
  COMPLEX               :: kIn( maxM ),  kOut( maxM ), T, kInt
  COMPLEX               :: sumk( MaxM )
  REAL (KIND=8), ALLOCATABLE :: kz2( : ), thetaT( : ), S( : )
  REAL         (KIND=8) :: omega, c0

  ! Open file for eigenfunctions                                  
  ! OPEN ( FILE = 'ZBarFile', UNIT = ZBarFile, STATUS = 'UNKNOWN' ) 

  P       = 0    ! zero out the pressure field
  delta_r = ( RmaxM - RminM ) / ( Nr - 1 )

  !  *** Loop over angle ***                                           

  Bearing: DO Itheta = 1, Ntheta 
     tsx = COS( DegRad * theta( Itheta ) )
     tsy = SIN( DegRad * theta( Itheta ) ) 

     ! Get modal values
     iElement = IElementSource 
     ! WRITE( *, * ) 'Tracing bearing ', Itheta, ATAN2( tsy, tsx ) / DegRad

     CALL SourceElement( iElement, Outside, RIn, ROut, xs, ys, tsx, tsy, MProp, M, maxM, &
          k, PhiR, PhiS, const, kIn, PhiIn, kOut, PhiOut )
     MProp = MIN( MLimit, MProp ) 

     IF ( MProp > 0 ) THEN   ! any propagating modes at the source?
        ! write modes at first range                                 
        ! IF ( Itheta == 1 ) THEN 
        !    WRITE( ZBarFile, * ) MProp 
        !    WRITE( ZBarFile, * ) const( 1 : MProp ) 
        ! END IF

        ! apply the source beam pattern
        ! using kIn for wavenumber; would be more precise to use interpolated value at source
        IF ( SBPFlag == '*' .AND. Itheta == 1 ) THEN
           ALLOCATE( kz2( MProp ), thetaT( MProp ), S( MProp ) )
           c0    = 1500   ! reference sound speed, should be speed at the source depth
           omega = 2 * pi * freq
           kz2   = REAL( omega ** 2 / c0 ** 2 - kIn( 1 : MProp ) ** 2 )      ! vertical wavenumber squared
           WHERE ( kz2 < 0 ) kz2 = 0                                         ! remove negative values

           thetaT = RadDeg * ATAN( sqrt( kz2 ) / REAL( kIn( 1 : MProp ) ) )  ! calculate the angle in degrees
           CALL interp1( SrcBmPat( :, 1 ), SrcBmPat( :, 2 ), thetaT, S )
           const( 1 : MProp ) = const( 1 : MProp ) * REAL( S )                       ! apply the shading
        END IF

        const( 1 : MProp ) = i * SQRT( 2.0 * pi ) * EXP( i * pi / 4.0 ) * const( 1 : MProp )
        sumk = 0.0

        ! *** March forward in range ***                                 

        RangeStepping: DO ir = 1, Nr 
           RM = RminM + ( ir - 1 ) * delta_r 
           IF ( RM == 0.0 ) RM = MIN( 1.0, delta_r )

           ! Crossing into new element?                                  
           EltLoop: DO WHILE ( RM > ROut ) 

              ! Copy outside info to inside                          
              NewElement         = AdjacentElement( Outside, iElement ) 
              RIn                =   ROut 
              kIn(   1 : MProp ) =   kOut( 1 : MProp ) 
              PhiIn( 1 : MProp ) = PhiOut( 1 : MProp ) 

              ! Get new outside info                                 
              CALL OUT( iElement, NewElement, Outside, ROut, xs, ys, tsx, tsy, MProp, M, maxM, k, PhiR, kOut, PhiOut )
              IF ( MProp <= 0 ) EXIT RangeStepping   ! jump out if there are no propagating modes
              iElement = NewElement 

           END DO EltLoop

           ! *** Compute modal contribution at this range ***            

           T = 0.0 

           IF ( RIn /= ROut ) THEN 
              alpha = ( RM - RIn ) / ( ROut - RIn ) 
              alpha = MIN( MAX( alpha,  0.0 ), 1.0 ) 
           ELSE 
              alpha = 0.0 
           ENDIF

           ! IF ( ir == Nr ) WRITE( ZBarFile, * ) MProp
           Mode: DO L = 1, MProp 
              kInt      =   kIn( L ) + alpha * (   kOut( L ) -   kIn( L ) ) 
              PhiInt    = PhiIn( L ) + alpha * ( PhiOut( L ) - PhiIn( L ) ) 
              sumk( L ) =  sumk( L ) + delta_r * kInt 
              T         = T + PhiInt * const( L ) * EXP( -i * sumk( L ) ) / SQRT( kInt )    
              ! IF ( ir == Nr ) WRITE( ZBarFile, * ) PhiInt  ! write mode at last range     
           END DO Mode

           P( Itheta, ir ) = CMPLX( T / SQRT( RM ) )

        END DO RangeStepping

        ! Write average wavenumber to file                           
        ! OPEN ( FILE = 'KBarFile', UNIT = KBarFile, STATUS = 'UNKNOWN' ) 
        ! WRITE( KBarFile, * ) MProp, RM 
        ! WRITE( KBarFile, * ) sumk( 1 : MProp ) / RM

     END IF
  END DO Bearing

  IF ( ALLOCATED( kz2 ) ) DEALLOCATE( kz2, thetaT, S )

END SUBROUTINE Evaluate3D

!**********************************************************************!

SUBROUTINE SourceElement( iElement, Outside, RIn, ROut, xs, ys, tsx, tsy, MProp, M, maxM, &
                          k, PhiR, PhiS, const, kIn, PhiIn, kOut, PhiOut )

  ! Given an element number for the source                            
  ! Computes modal information

  USE ElementMod
  IMPLICIT NONE
  INTEGER, INTENT( IN  ) :: M( * ), maxM                                ! number of modes
  REAL,    INTENT( IN  ) :: xs, ys, tsx, tsy                            ! source coordinates, radial tangent
  COMPLEX, INTENT( IN  ) :: PhiS( maxM, * ), PhiR( maxM, * )            ! mode shapes
  COMPLEX, INTENT( IN  ) :: k(   maxM, * )
  COMPLEX, INTENT( OUT ) :: kIn( maxM ),  kOut( maxM ) ! wavenumbers
  INTEGER, INTENT( OUT ) :: MProp                   ! number of modes propagating
  INTEGER, INTENT( OUT ) :: Outside                 ! side through which radial exits
  REAL,    INTENT( OUT ) :: RIn, ROut               ! in and out ranges for the radial
  COMPLEX, INTENT( OUT ) :: PhiIn( * ), PhiOut( * ) ! mode shapes
  COMPLEX, INTENT( OUT ) :: Const( * )              ! interpolated mode excitation coefficients
  INTEGER :: Inside, IBad, IGood1, IGood2, &
             iCorner1, iCorner2, iCorner3, iCorner4, iElement, &
             ISide, Iset1, Iset2, L, node1, node2
  REAL    :: R,  RV( 3 ), SV( 3 ), RVC( 3 ), SIn, SOut, x1S, y1S, xCenter, yCenter, &
             Delta, tx, ty, txC, tyC, alpha

  MProp = HUGE( MProp )

  ! Coordinates of the centroid of the source element                 

  xCenter = SUM( x( node( 1 : 3, iElement ) ) ) / 3.0
  yCenter = SUM( y( node( 1 : 3, iElement ) ) ) / 3.0 

  Side: DO ISide = 1, 3 

     node1 = node( iCorner( ISide, 1 ), iElement )
     node2 = node( iCorner( ISide, 2 ), iElement )
     Iset1 = Iset( node1 )
     Iset2 = Iset( node2 ) 
     MProp = MIN( MProp, M( Iset1 ), M( Iset2 ) ) 

     x1S = x( node1 ) - xs
     y1S = y( node1 ) - ys
     txC = x( node1 ) - xCenter
     tyC = y( node1 ) - yCenter                                 
     tx  = x( node2 ) - x( node1 )
     ty  = y( node2 ) - y( node1 ) 

     Delta = tsx * ty - tsy * tx 

     ! *** Radial parallel to side? ***                               

     IF ( Delta == 0.0 ) THEN 
        SV(  ISide ) = HUGE( SV( ISide ) )
     ELSE 
        RVC( ISide ) = ( txC * ty  - tyC * tx  ) / Delta 
        RV(  ISide ) = ( x1S * ty  - y1S * tx  ) / Delta 
        SV(  ISide ) = ( x1S * tsy - y1S * tsx ) / Delta 
     ENDIF
  END DO Side

  ! Identify two good sides and one bad side based on the intercept point

  IBad = 1 
  IF ( ABS( SV( 2 ) - 0.5 ) > ABS( SV( IBad ) - 0.5 ) ) IBad = 2                                                       
  IF ( ABS( SV( 3 ) - 0.5 ) > ABS( SV( IBad ) - 0.5 ) ) IBad = 3                                                     

  IGood1 = 1
  IGood2 = 2
  IF ( IBad == 1 ) THEN 
     IGood1 = 3 
  ELSE IF ( IBad == 2 ) THEN 
     IGood2 = 3 
  ENDIF

  ! The side with the lesser RVC is the inside                    

  IF ( RVC( IGood1 ) < RVC( IGood2 ) ) THEN 
     Inside  = IGood1
     Outside = IGood2 
  ELSE 
     Inside  = IGood2
     Outside = IGood1 
  ENDIF

  sIn      = SV( Inside ) 
  sIn      = MIN( MAX( sIn,  0.0 ), 1.0 ) 
  RIn      = RV( Inside ) 

  SOut     = SV( Outside ) 
  SOut     = MIN( MAX( SOut, 0.0 ), 1.0 ) 
  ROut     = RV( Outside ) 

  ! Get values of modes at Z = SD and (x, y) = intercept points   

  iCorner1 = iCorner(  Inside, 1 )
  iCorner2 = iCorner(  Inside, 2 ) 
  iCorner3 = iCorner( Outside, 1 )
  iCorner4 = iCorner( Outside, 2 ) 

  ! Interpolate to get modal values at source                     
  R = 0.0 

  IF ( RIn /= ROut ) THEN 
     alpha = ( R - RIn ) / ( ROut - RIn ) 
     alpha = MIN( MAX( alpha,  0.0 ), 1.0 ) 
  ELSE 
     alpha = 0.0 
  ENDIF

  Mode: DO L = 1, MProp 
     PhiIn(  L ) = PhiS( L, iCorner1 ) + sIn   * ( PhiS( L, iCorner2 ) - PhiS( L, iCorner1 ) )   
     PhiOut( L ) = PhiS( L, iCorner3 ) + SOut  * ( PhiS( L, iCorner4 ) - PhiS( L, iCorner3 ) )   
     const(  L ) = PhiIn( L )          + alpha * ( PhiOut( L )         - PhiIn( L ) ) 
  END DO Mode

  ! Obtain values of modes at z = Rd and (x, y)=intercept points      

  CALL InterpolateModes( iElement, Inside,  sIn,  MProp, M, maxM, k, PhiR, kIn,  PhiIn  )          
  CALL InterpolateModes( iElement, Outside, sOut, MProp, M, maxM, k, PhiR, kOut, PhiOut )          
 
END SUBROUTINE SourceElement

!**********************************************************************!

SUBROUTINE OUT( iElement, NewElement, Outside, ROut, xs, ys, tsx, tsy, MProp, M, maxM, k, PhiR, kOut, PhiOut )

  ! Given an element number and a side through which prop path enters
  ! computes mode info at exit point

  USE ElementMod
  IMPLICIT NONE
  INTEGER, INTENT( IN  ) :: NewElement
  INTEGER, INTENT( IN  ) :: iElement         ! Current element number
  INTEGER, INTENT( IN  ) :: M( * ), maxM     ! number of modes, max number allowed
  REAL,    INTENT( IN  ) :: xs, ys           ! source coordinate
  REAL,    INTENT( IN  ) :: tsx, tsy         ! tangent for the bearing line
  COMPLEX, INTENT( IN  ) :: PhiR( maxM, * )  ! mode shapes at each node
  INTEGER, INTENT( OUT ) :: Outside          ! side through which path exits
  INTEGER, INTENT( OUT ) :: MProp            ! number of propagating modes
  REAL,    INTENT( OUT ) :: ROut             ! range at which path exits
  COMPLEX, INTENT( OUT ) :: kOut( maxM ), PhiOut( * )   ! modes at exit point
  INTEGER :: ISide, node1, node2
  REAL    :: ROutT, sOut, sT, x1S, y1S, Delta, Tx, Ty
  COMPLEX :: k(    maxM, * )

  ! WRITE( *, * ) '   Crossing into new element = ', NewElement           

  ! If no adj elt then exit. Previous values of KOut, PhiOut retained 

  IF ( NewElement == 0 ) THEN 
     ROut = HUGE( ROut )
     RETURN 
  ENDIF

  ! *** Loop over sides to find outside ***                           

  sOut = HUGE( sOut )

  Side: DO ISide = 1, 3 

     ! Don't try an outside the same as the inside             
     IF ( AdjacentElement( ISide, NewElement ) /= iElement ) THEN 

        node1 = node( iCorner( ISide, 1 ), NewElement ) 
        node2 = node( iCorner( ISide, 2 ), NewElement ) 

        x1S = x( node1 ) - xs
        y1S = y( node1 ) - ys                                               
        Tx  = x( node2 ) - x( node1 )
        Ty  = y( node2 ) - y( node1 ) 

        Delta = tsx * Ty - tsy * Tx 

        ! *** Radial parallel to side? ***                            

        IF ( Delta == 0.0 ) THEN 
           sT    = HUGE( sT )
        ELSE 
           ROutT = ( x1S * Ty  - y1S * Tx  ) / Delta 
           sT    = ( x1S * tsy - y1S * tsx ) / Delta 
        ENDIF
        ! If intercept is in segment, store the side number       
        IF ( ABS( sT - 0.5 ) < ABS( sOut - 0.5) ) THEN 
           Outside = ISide 
           sOut    = sT 
           ROut    = ROutT
           ! write( *, * ) 'nodes', ModeFileName( Node1 ), ModeFileName( Node2 )
        ENDIF
     ENDIF
  END DO Side

  ! Obtain values of modes at intercept points                        

  CALL InterpolateModes( NewElement, Outside, sOut, MProp, M, maxM, k, PhiR, kOut, PhiOut )          
 
END SUBROUTINE OUT

!**********************************************************************!

SUBROUTINE InterpolateModes( iElement, ISide, s, MProp, M, maxM, k, PhiR, kInt, PhiInt )          

  ! Given an element, a side, and a proportional distance along the side                                       
  ! Returns interpolated modal values

  USE ElementMod
  IMPLICIT NONE
  INTEGER, INTENT( IN  ) :: iElement, ISide  ! index of the element and side
  INTEGER, INTENT( IN  ) :: M( * ), maxM     ! number of modes, max number allowed
  REAL,    INTENT( IN  ) :: s                ! proportional distance along the side
  COMPLEX, INTENT( IN  ) :: PhiR( maxM, * )  ! mode shapes at each node
  COMPLEX, INTENT( IN  ) :: k(   maxM, * )   ! wavenumbers at each node
  INTEGER, INTENT( OUT ) :: MProp            ! number of propagating modes
  COMPLEX, INTENT( OUT ) :: PhiInt( * )      ! interpolated mode shapes
  COMPLEX, INTENT( OUT ) :: kInt( maxM )     ! interpolated wavenumbers
  INTEGER                :: Iset1, Iset2, I, node1, node2 
  REAL                   :: st

  node1 = node( iCorner( ISide, 1 ), iElement )
  node2 = node( iCorner( ISide, 2 ), iElement )                                         
  Iset1 = Iset( node1 )
  Iset2 = Iset( node2 ) 
  MProp = MIN( MProp, M( Iset1 ), M( Iset2 ) ) 

  st = MIN( MAX( s, 0.0 ), 1.0 ) ! Extrapolation blocked by making sure s in [0, 1]

  Mode: DO I = 1, MProp 
     kInt(   I ) =    k( I, Iset1 ) + st * (    k( I, Iset2 ) -    k( I, Iset1 ) )  
     PhiInt( I ) = PhiR( I, Iset1 ) + st * ( PhiR( I, Iset2 ) - PhiR( I, Iset1 ) )  
  END DO Mode

END SUBROUTINE InterpolateModes
